/* StegFS - The Steganographic File System */
/* Copyright (c)1999 Andrew McDonald <andrew@mcdonald.org.uk> */

/* Take an ext2 fs and make it ready for data hiding */
/* - Randomize contents of unused blocks
   - Create root inode(s)
   - Create stegfs information table
*/
/* $Id: mkstegfs.c,v 1.11 1999/08/06 16:07:25 adm36 Exp adm36 $ */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include "stegfs.h"
#include "mkstegfs.h"
#include <sys/time.h>
#include <sys/stat.h>
#include <string.h>
#include "sha1.h"
#include "serpent.h"


int main(int argc, char* args[]) {

  struct stegfs_fs fs;
  struct stegfs_fs *fssys = &fs;
  u_int32_t i,j;
  char *passwds[16];
  char *skeys[16];
  FILE *devrand;

  char *defcipher = "serpent";
  char *cipherstr;

  int levels = 0;
  char promptbuf[35];
  char *pass;

  printf("StegFS - (c)1999 ADM\n");
  if(argc < 4 || argc > 5) {
    printusage();
    return 1;
  }

  levels = atoi(args[3]);
  if (levels<1 || levels>15) {
    printf("Invalid number of levels\n");
    return 2;
  }

  if (argc<5)
    cipherstr = defcipher;
  else
    cipherstr = args[4];

  if (!open_fs(fssys, args[1], args[2], cipherstr)) {
    printf("Failed to open device or file\n");
    return 2;
  }
  
  printf("cipher: %s\n", fssys->cipher->id);
  if(!get_ext2_super(fssys)) {
    return 3;
  }

  /* get passphrases */
  for (i=0; i<levels; i++) {
    sprintf(promptbuf, "Enter passphrase for level %u: ", i+1);
    pass = getpass(promptbuf);

    passwds[i] = malloc(strlen(pass)+1);
    strncpy(passwds[i], pass, strlen(pass)+1);

    sprintf(promptbuf, "Re-enter passphrase for level %u: ", i+1);
    pass = getpass(promptbuf);
    if (strcmp(pass,passwds[i]) != 0) {
      printf("Passphrases differ.\n");
      i--;
    }
  }
  passwds[i] = NULL;

  initbtab(fssys);

  get_bbpos(fssys);

  /*
  for(i=0; i<fssys->groups; i++) {
    printf("%u ", fssys->bbarray[i]);
  }
  printf("\n");
  */
  randomize_empties(fssys);
  /*
  printf("1 %d\n", isblockused(fssys, 8191));
  printf("2 %d\n", isblockused(fssys, 8192));
  printf("3 %d\n", isblockused(fssys, 8193));
  printf("4 %d\n", isblockused(fssys, 8194));
  */

  devrand = fopen("/dev/urandom", "r");
  if(devrand==NULL) {
    printf("Error opening /dev/urandom\n");
    exit(99);
  }
  fseek(fssys->btabfile,
	fssys->sb.s_blocks_count*sizeof(struct stegfs_btable), SEEK_SET);
  for (i=0; i<levels; i++) {
    char enckey[16];
    unsigned int *xkey;
    skeys[i] = malloc(16);
    fread(skeys[i], 16, 1, devrand);
    xkey = (unsigned int *)skeys[i];
    /*
    printf("level: %u key: %08x %08x %08x %08x\n", i+1,
	   xkey[0], xkey[1], xkey[2], xkey[3]);
    */
    for (j=0; j<(i+1); j++) {
      encrypt_skey(fssys, passwds[i], skeys[j], enckey);
      fwrite(enckey, sizeof(enckey), 1, fssys->btabfile);
    }
  }
  fclose(devrand);

  for (i=0; i<levels; i++) {
    mkroot(fssys, i+1, passwds[i], skeys);
  }

  close_fs(fssys);

  return 0;
}




void printusage() {
  printf("\nUsage:\n");
  printf("mkstegfs fsdevice btabfile num_levels cipher\n");

  return;
}



void initbtab(struct stegfs_fs *fs) {

  int i,j;
  struct stegfs_btable btab;
  FILE *devrand;
  char key[16];

#ifdef DEVEL_ZEROFILL
  devrand = fopen("/dev/zero", "r");
#else
  devrand = fopen("/dev/urandom", "r");
#endif

  if(devrand==NULL) {
    printf("Error opening /dev/urandom\n");
    return;
  }

  /* create the stegfstab file */
  srandom(getrand32());
  printf("Writing block table: ");
  for(i=0; i<fs->sb.s_blocks_count; i++) {
    printf("%3d%%", (100*i)/fs->sb.s_blocks_count);

    fread(&btab, sizeof(struct stegfs_btable), 1, devrand);

    fwrite(&btab, sizeof(struct stegfs_btable), 1, fs->btabfile);
    printf("\b\b\b\b");
  }

  for (i=1; i<STEGFS_MAX_LEVELS; i++) {
    for (j=0; j<i; j++) {
      fread(&key, sizeof(key), 1, devrand);
      fwrite(&key, sizeof(key), 1, fs->btabfile);
    }
  }
  printf("Done\n");
  fclose(devrand);
  return;

}


void mkroot(struct stegfs_fs *fs, int level, char *passwd, char *skeys[]) {

  u_int32_t inonum;
  int i;
  struct timeval tv;
  struct timezone tz;
  struct stegfs_btable btab, btab2;
  struct stegfs_inode inode, inode2;
  char *datablock, *datablock2, *key, *keys;
  struct ext2_dir_entry_2 *dirent;
  u_int32_t *iblks;

  inonum = EXT2_ROOT_INO | STEGFS_INO_MASK | (level)<<(STEGFS_LVL_SHIFT);

  btab.magic1 = 0;
  btab.magic2 = 1;
  btab.ino = inonum;

  inode.i_mode = S_IFDIR|S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH;
  inode.i_uid = fs->sb.s_def_resuid;
  inode.i_size = EXT2_BLOCK_SIZE(&(fs->sb));
  gettimeofday(&tv, &tz);
  inode.i_atime = inode.i_ctime = inode.i_mtime = tv.tv_sec;
  inode.i_dtime = 0;
  inode.i_gid = fs->sb.s_def_resgid;
  inode.i_links_count = 3; /* /, . and .. */
  inode.i_blocks = 1;
  inode.i_flags = 0;
  inode.osd1.linux1.l_i_reserved1 = 0;

  inode.i_version = 0;
  inode.i_file_acl = 0;
  inode.i_dir_acl = 0;
  inode.i_faddr = 0;
  inode.osd2.linux2.l_i_frag = 0;
  inode.osd2.linux2.l_i_fsize = 0;
  inode.osd2.linux2.i_pad1 = 0;
  inode.osd2.linux2.l_i_reserved2[0] = inode.osd2.linux2.l_i_reserved2[1] = 0;

  inode.i_icopies = NCOPIES;
  inode.i_bcopies = NCOPIES;
  inode.i_pad0 = 0;

  for(i=0; i<EXT2_N_BLOCKS*STEGFS_MAX_BLOCK_COPIES+STEGFS_MAX_INO_COPIES; i++)
    inode.i_inode[i] = 0;

  iblks = malloc(inode.i_icopies * sizeof(u_int32_t));
  get_new_inode(fs, inonum, inode.i_icopies, iblks, skeys[level-1]);

  for(i=0; i<inode.i_icopies; i++) {
    inode.i_inode[i] = iblks[i];
    key = make_key(skeys[level-1], inode.i_inode[i]);
    keys = fs->cipher->set_key(key, 128);
    free(key);
    fs->cipher->encrypt(keys, (char *)&btab, (char *)&btab2);
    free(keys);
    put_btab(fs, inode.i_inode[i], &btab2);
    printf("i %u\n", inode.i_inode[i]);
  }
  free(iblks);

  btab.magic2 = 0;
  for(i=0; i<inode.i_bcopies; i++) {
    inode.i_block[i*EXT2_N_BLOCKS] = get_new_block(fs);
    key = make_key(skeys[level-1], inode.i_block[i*EXT2_N_BLOCKS]);
    keys = fs->cipher->set_key(key, 128);
    free(key);
    fs->cipher->encrypt(keys, (char *)&btab, (char *)&btab2);
    free(keys);
    put_btab(fs, inode.i_block[i*EXT2_N_BLOCKS], &btab2);
    printf("d %u\n", inode.i_block[i*EXT2_N_BLOCKS]);
  }

  btab.magic2 = 1;

  for(i=0; i<inode.i_icopies; i++) {
    key = make_key(skeys[level-1], inode.i_inode[i]);
    encrypt_cbc2(fs, key, (char *)&inode, (char *)&inode2,
		 sizeof(struct stegfs_inode));
    fseek(fs->fsfile, EXT2_BLOCK_SIZE(&(fs->sb))*inode.i_inode[i], SEEK_SET);
    fwrite(&inode2, sizeof(struct stegfs_inode), 1, fs->fsfile);
    btab.bchecksum = stegfs_chksum((unsigned char *)&inode2,
				   sizeof(struct stegfs_inode));
    keys = fs->cipher->set_key(key, 128);
    fs->cipher->encrypt(keys, (char *)&btab, (char *)&btab2);
    free(key);
    free(keys);
    put_btab(fs, inode.i_inode[i], &btab2);
  }

  datablock = malloc( EXT2_BLOCK_SIZE(&(fs->sb)) );
  datablock2 = malloc( EXT2_BLOCK_SIZE(&(fs->sb)) );
  dirent = (struct ext2_dir_entry_2 *)
    malloc(sizeof(struct ext2_dir_entry_2) * 2);
  dirent[0].rec_len = 12;
  dirent[0].name_len = 1;
  dirent[0].file_type = 0;
  strncpy(dirent[0].name, ".", EXT2_NAME_LEN);
  dirent[0].inode = inonum;

  dirent[1].rec_len = EXT2_BLOCK_SIZE(&(fs->sb)) - dirent[0].rec_len;
  dirent[1].name_len = 2;
  dirent[1].file_type = 0;
  strncpy(dirent[1].name, "..", EXT2_NAME_LEN);
  dirent[1].inode = inonum;

  memcpy(datablock, &dirent[0], sizeof(struct ext2_dir_entry_2));
  memcpy(datablock+dirent[0].rec_len, &dirent[1],
	 sizeof(struct ext2_dir_entry_2));

  btab.magic2 = 0;
  for(i=0; i<inode.i_bcopies; i++) {
    key = make_key(skeys[level-1], inode.i_block[i*EXT2_N_BLOCKS]);
    fseek(fs->fsfile,
	  EXT2_BLOCK_SIZE(&(fs->sb))*inode.i_block[i*EXT2_N_BLOCKS],
	  SEEK_SET);
    encrypt_cbc2(fs, key, datablock, datablock2,
		 EXT2_BLOCK_SIZE(&(fs->sb)));
    fwrite(datablock2, EXT2_BLOCK_SIZE(&(fs->sb)), 1, fs->fsfile);
    btab.bchecksum = stegfs_chksum(datablock2, EXT2_BLOCK_SIZE(&(fs->sb)));
    keys = fs->cipher->set_key(key, 128);
    fs->cipher->encrypt(keys, (char *)&btab, (char *)&btab2);
    free(key);
    free(keys);
    put_btab(fs, inode.i_block[i*EXT2_N_BLOCKS], &btab2);

  }

  free(dirent);
  free(datablock);
  free(datablock2);
}



u_int32_t getrand32() {
  FILE *devrand;
  u_int32_t retval;

  devrand = fopen("/dev/urandom", "r");
  if(devrand==NULL) {
    printf("Error opening /dev/urandom\n");
    exit(101);
  }

  fread(&retval, sizeof(u_int32_t), 1, devrand);

  fclose(devrand);

  return retval;

}

void getrand(void *data, u_int32_t len) {
  FILE *devrand;
  devrand = fopen("/dev/urandom", "r");
    if(devrand==NULL) {
    printf("Error opening /dev/urandom\n");
    exit(101);
  }

  fread(data, len, 1, devrand);

  fclose(devrand);

  return;
}

int u32comp(const void *a, const void *b) {
  u_int32_t *aa, *bb;

  aa = (u_int32_t *)a;
  bb = (u_int32_t *)b;

  if(*aa > *bb)
    return 1;
  if(*aa < *bb)
    return -1;
  
  return 0;
}

int blockinlist(u_int32_t b, u_int32_t *list, u_int32_t i) {
  u_int32_t j;
  for(j=0; j<i; j++) {
    if(list[j]==b)
      return 1;
  }
  return 0;
}


